"""Traces and packs notebook environments with ReproZip.
"""

__version__ = '0.4'


def _jupyter_nbextension_paths():
    return [
        dict(
            section='notebook',
            src='notebook-extension.js',
            dest='reprozip-jupyter.js',
            require='reprozip-jupyter',
        ),
    ]


def _jupyter_server_extension_paths():
    return [
        dict(module='reprozip_jupyter.server_extension'),
    ]



#!C:\bld\reprozip-jupyter_1526430084550\_b_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'reprozip-jupyter==0.4','console_scripts','reprozip-jupyter'
__requires__ = 'reprozip-jupyter==0.4'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('reprozip-jupyter==0.4', 'console_scripts', 'reprozip-jupyter')()
    )



from datetime import datetime
from notebook.utils import url_path_join as ujoin
from rpaths import Path
import subprocess
import sys
from tornado.process import Subprocess
from tornado.web import RequestHandler, asynchronous


try:
    unicode_ = unicode
except NameError:
    unicode_ = str


class TraceHandler(RequestHandler):
    def initialize(self, nbapp=None):
        self.nbapp = nbapp
        self._tempdir = None

    @asynchronous
    def post(self):
        self._notebook_file = Path(self.get_body_argument('file'))
        name = self._notebook_file.unicodename
        if name.endswith('.ipynb'):
            name = name[:-6]
        name = u'%s_%s.rpz' % (name, datetime.now().strftime('%Y%m%d-%H%M%S'))
        self._pack_file = self._notebook_file.parent / name
        self.nbapp.log.info("reprozip: tracing request from client: file=%r",
                            self._notebook_file)
        self._tempdir = Path.tempdir()
        self.nbapp.log.info("reprozip: created temp directory %r",
                            self._tempdir)
        proc = Subprocess(
            [sys.executable, '-c',
             'from reprozip_jupyter.main import main; main()',
             'trace',
             '--dont-save-notebook',
             '-d', self._tempdir.path,
             self._notebook_file.path],
            stdin=subprocess.PIPE)
        proc.stdin.close()
        proc.set_exit_callback(self._trace_done)
        self.nbapp.log.info("reprozip: started tracing...")

    def _trace_done(self, returncode):
        self.nbapp.log.info("reprozip: tracing done, returned %d", returncode)
        if returncode == 0:
            # Pack
            if self._pack_file.exists():
                self._pack_file.remove()
            proc = Subprocess(
                ['reprozip', 'pack', '-d',
                 self._tempdir.path,
                 self._pack_file.path],
                stdin=subprocess.PIPE)
            proc.stdin.close()
            proc.set_exit_callback(self._packing_done)
            self.nbapp.log.info("reprozip: started packing...")
        else:
            self._tempdir.rmtree()
            if returncode == 3:
                self.set_header('Content-Type', 'application/json')
                self.finish(
                    {'error': "There was an error running the notebook. "
                              "Please make sure that it can run from top to "
                              "bottom without error before packing."})
            else:
                self.send_error(500)

    def _packing_done(self, returncode):
        self.nbapp.log.info("reprozip: packing done, returned %d", returncode)
        if returncode == 0:
            # Send the response
            self.set_header('Content-Type', 'application/json')
            self.finish({'bundle': unicode_(self._pack_file)})
            self.nbapp.log.info("reprozip: response sent!")
        else:
            self.send_error(500)
        self._tempdir.rmtree()


def load_jupyter_server_extension(nbapp):
    nbapp.log.info('reprozip: notebook extension loaded')

    webapp = nbapp.web_app
    base_url = webapp.settings['base_url']
    webapp.add_handlers(".*$", [
        (ujoin(base_url, r"/reprozip/trace"), TraceHandler, {'nbapp': nbapp}),
    ])



# Copyright (C) 2014-2017 New York University
# This file is part of ReproZip which is released under the Revised BSD License
# See file LICENSE for full license details.

"""Trace a notebook to generate accompanying RPZ pack.
"""

from __future__ import division, print_function, unicode_literals

import argparse
from jupyter_client.launcher import launch_kernel
from jupyter_client.manager import KernelManager
from jupyter_client.managerabc import KernelManagerABC
import logging
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors.execute import CellExecutionError
import nbformat
import os
from rpaths import Path
import sys

from reprounzip.common import setup_logging


logger = logging.getLogger('reprozip_jupyter')


class RPZOptions(object):
    def __init__(self, verbosity=1, dir=None, identify_packages=True,
                 find_inputs_outputs=True, append=None):
        self.verbosity = verbosity
        self.dir = dir
        self.identify_packages = identify_packages
        self.find_inputs_outputs = find_inputs_outputs
        self.append = append

    def trace_command_line(self, kernel_cmd):
        cmd = ['reprozip']
        cmd.extend(['-v'] * (self.verbosity - 1))
        cmd.append('trace')
        if self.dir:
            cmd.extend(['--dir', self.dir])
        if not self.identify_packages:
            cmd.append('--dont-identify-packages')
        if not self.find_inputs_outputs:
            cmd.append('--dont-find-inputs-outputs')
        if self.append:
            cmd.append('--continue')
        else:
            cmd.append('--overwrite')
        cmd.extend(kernel_cmd)
        return cmd

    def config_file(self):
        if self.dir:
            return Path(self.dir) / 'config.yml'
        else:
            return Path('.reprozip-trace/config.yml')


class RPZKernelManager(KernelManager):
    rpz_options = None

    def _launch_kernel(self, kernel_cmd, **kw):
        cmd = self.rpz_options.trace_command_line(kernel_cmd)

        logger.info("Kernel requested, connection file: %s",
                    self.connection_file)
        logger.info("Executing: %r", cmd)
        return launch_kernel(cmd, **kw)

    def finish_shutdown(self, *args, **kwargs):
        kwargs['waittime'] = 600
        super(RPZKernelManager, self).finish_shutdown(*args, **kwargs)

        # Add the input file to the configuration
        config = self.rpz_options.config_file()

        with config.rewrite(encoding='utf-8') as (read, write):
            for line in read:
                write.write(line)
                if line == 'inputs_outputs:\n':
                    write.write('  - name: jupyter_connection_file'
                                '  # Needed for reprozip-jupyter operations\n'
                                '    read_by_runs: [0]\n'
                                '    path: %s\n' % self.connection_file)


KernelManagerABC.register(RPZKernelManager)


class RPZExecutePreprocessor(ExecutePreprocessor):
    def __init__(self, options):
        self.rpz_options = options
        super(RPZExecutePreprocessor, self).__init__()

    def preprocess(self, nb, resources):
        # no change {
        path = resources.get('metadata', {}).get('path', '')
        if path == '':
            path = None

        kernel_name = nb.metadata.get('kernelspec', {}).get('name', 'python')
        if self.kernel_name:
            kernel_name = self.kernel_name
        self.log.info("Executing notebook with kernel: %s" % kernel_name)
        # } no change

        logger.info("Starting kernel...")

        # copied from start_new_kernel(), but using our KernelManager class {
        km = RPZKernelManager(kernel_name=kernel_name)
        km.rpz_options = self.rpz_options
        km.start_kernel(extra_arguments=self.extra_arguments,
                        cwd=path)  # changed not to hide stderr
        kc = km.client()
        kc.start_channels()
        try:
            kc.wait_for_ready(timeout=60)
        except RuntimeError:
            kc.stop_channels()
            km.shutdown_kernel()
            raise
        # } start_new_kernel()

        self.km, self.kc = km, kc
        logger.info("Kernel started")

        # no change {
        self.kc.allow_stdin = False

        try:
            nb, resources = super(ExecutePreprocessor, self).preprocess(
                nb, resources)
        except CellExecutionError:
            sys.exit(3)
        finally:
            self.kc.stop_channels()
            self.km.shutdown_kernel(now=False)  # changed from now=False

        return nb, resources
        # } no change


def trace_notebook(filename, save_notebook=True, **kwargs):
    with open(filename) as fp:
        notebook = nbformat.read(fp, as_version=4)
    preprocessor = RPZExecutePreprocessor(RPZOptions(**kwargs))
    preprocessor.preprocess(
        notebook,
        {'metadata': {'path': os.path.dirname(filename)}})
    if save_notebook:
        with open(filename, 'wt') as fp:
            nbformat.write(notebook, fp)
    return notebook


def cmd_trace_notebook(args):
    setup_logging('REPROZIP-JUPYTER-TRACE', args.verbosity)
    if not args.notebook:
        sys.stderr.write("missing notebook\n")
        sys.exit(2)
    return trace_notebook(args.notebook,
                          save_notebook=args.save_notebook,
                          verbosity=args.verbosity, dir=args.dir,
                          identify_packages=args.identify_packages,
                          find_inputs_outputs=args.find_inputs_outputs,
                          append=args.append)


def setup(parser):
    parser.add_argument('-d', '--dir',
                        help="where to store database and configuration file "
                             "(default: ./.reprozip-trace)")
    parser.add_argument(
        '--dont-save-notebook', action='store_false', default=True,
        dest='save_notebook',
        help="do not update the notebook file when executing")
    parser.add_argument(
        '--dont-identify-packages', action='store_false', default=True,
        dest='identify_packages',
        help="do not try identify which package each file comes from")
    parser.add_argument(
        '--find-inputs-outputs', action='store_true',
        default=False, dest='find_inputs_outputs',
        help="try to identify input and output files")
    parser.add_argument(
        '--dont-find-inputs-outputs', action='store_false',
        default=False, dest='find_inputs_outputs',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '-c', '--continue', action='store_true', dest='append',
        help="add to the previous trace, don't replace it")
    parser.add_argument(
        '-w', '--overwrite', action='store_true', dest='overwrite',
        help="overwrite the previous trace, don't add to it")
    parser.add_argument('notebook', help="command-line to run under trace")
    parser.set_defaults(func=cmd_trace_notebook)



# Copyright (C) 2014-2017 New York University
# This file is part of ReproZip which is released under the Revised BSD License
# See file LICENSE for full license details.

import argparse
import sys

from . import __version__
from . import run
from . import trace


def main():
    def add_options(opts):
        opts.add_argument('-v', '--verbose', action='count', default=0,
                          dest='verbosity', help="augments verbosity level")
        opts.add_argument('--version', action='version',
                          version="reprounzip-jupyter version %s" %
                                  __version__)

    parser = argparse.ArgumentParser(
        description="Jupyter Notebook tracing/reproduction for ReproZip",
        epilog="Please report issues to reprozip-users@vgc.poly.edu")
    add_options(parser)
    subparser = parser.add_subparsers(title="subcommands", metavar='',
                                      dest='cmd')

    parser_trace = subparser.add_parser(
        'trace',
        help="Runs a Jupyter notebook under ReproZip trace to generate the "
             "accompanying environment package")
    add_options(parser_trace)
    trace.setup(parser_trace)

    parser_run = subparser.add_parser(
        'run',
        help="Runs a Jupyter notebook server that will spawn notebooks in "
             "Docker containers running in the given unpacked environment")
    add_options(parser_run)
    run.setup(parser_run)

    args = parser.parse_args()
    if getattr(args, 'func', None) is None:
        parser.print_help(sys.stderr)
        sys.exit(2)
    args.func(args)



# Copyright (C) 2014-2017 New York University
# This file is part of ReproZip which is released under the Revised BSD License
# See file LICENSE for full license details.

"""Run notebooks in a packed environment.
"""

from __future__ import division, print_function, unicode_literals

import argparse
import contextlib
import json
from jupyter_client.ioloop import IOLoopKernelManager
import logging
from notebook.notebookapp import NotebookApp
from notebook.services.kernels.kernelmanager import MappingKernelManager
import os
from rpaths import Path
import subprocess
import sys

from reprounzip.common import setup_logging


logger = logging.getLogger('reprozip_jupyter')


@contextlib.contextmanager
def process_connection_file(original):
    with original.open('rb') as fp:
        data = json.load(fp)

    data['ip'] = '0.0.0.0'  # Kernel should listen on all interfaces

    ports = [value for key, value in data.items() if key.endswith('_port')]

    fd, fixed_file = Path.tempfile(suffix='.json')
    with fixed_file.open('wb') as fp:
        json.dump(data, fp)
    os.close(fd)

    yield fixed_file, ports

    fixed_file.remove()


class RPZKernelManager(IOLoopKernelManager):
    rpz_target = None
    rpz_verbosity = 1

    def _launch_kernel(self, kernel_cmd, **kw):
        # Need to parse kernel command-line to find the connection file
        logger.info("Kernel command-line: %s", ' '.join(kernel_cmd))
        connection_file = None
        for i, arg in enumerate(kernel_cmd):
            if arg == '-f':
                connection_file = Path(kernel_cmd[i + 1])
                break

        if connection_file is None:
            logger.critical("The notebook didn't pass a connection file to "
                            "the kernel")
            sys.exit(1)

        with process_connection_file(connection_file) as (fixed_file, ports):
            # Upload connection file to environment
            subprocess.check_call(
                ['reprounzip'] + (['-v'] * (self.rpz_verbosity - 1)) +
                ['docker', 'upload', self.rpz_target,
                 '%s:jupyter_connection_file' % fixed_file])

        docker_options = []
        for port in ports:
            docker_options.extend(['-p', '%d:%d' % (port, port)])

        return subprocess.Popen(
            ['reprounzip'] + (['-v'] * (self.rpz_verbosity - 1)) +
            ['docker', 'run'] +
            ['--docker-option=%s' % opt for opt in docker_options] +
            [self.rpz_target])


class RPZMappingKernelManager(MappingKernelManager):
    def __init__(self, **kwargs):
        kwargs['kernel_manager_class'] = \
            'reprozip_jupyter.run.RPZKernelManager'
        super(RPZMappingKernelManager, self).__init__(**kwargs)


def run_server(target, jupyter_args=None, verbosity=1):
    RPZKernelManager.rpz_target = target
    RPZKernelManager.rpz_verbosity = verbosity

    if not jupyter_args:
        jupyter_args = []

    logger.info("Starting Jupyter notebook")
    NotebookApp.launch_instance(argv=jupyter_args,
                                kernel_manager_class=RPZMappingKernelManager)


def cmd_run_server(args):
    setup_logging('REPROZIP-JUPYTER-SERVER', args.verbosity)
    if not args.target:
        sys.stderr.write("Missing experiment directory\n")
        sys.exit(2)

    run_server(args.target, args.jupyter_args, verbosity=args.verbosity)


def setup(parser):
    parser.add_argument('target', help="Experiment directory")
    parser.add_argument('jupyter_args', nargs=argparse.REMAINDER,
                        help="Arguments to pass to the notebook server")
    parser.set_defaults(func=cmd_run_server)
